/**
 * @file
 * Provides functions to subscribe to data, notifications and api from the HAM DEX ESB.
 * 
 * @author
 * Daniel
 *
 * @param {object} app Express app.
 * @param {object} options Options.
 * @param {string} urlPrefix Prefix of the url. (optional)
 * @returns {object} Functions to get the inventory, subscribe to notifications and call api's.
 */
module.exports = function (app, options, urlPrefix) {

    const request = require('request').defaults({ rejectUnauthorized: false });

    // Determine IP address for callback, if not set
    if ((typeof options.callback !== 'undefined') && (typeof options.callback.host === 'undefined')) {
        const os = require('os');
        const addresses = Object.values(os.networkInterfaces()).reduce((r, list) => r.concat(list.reduce((rr, i) => rr.concat(i.family === 'IPv4' && !i.internal && i.address || []), [])), []);
        if (addresses.length > 0) {
            options.callback.host = addresses[0];
        }
    }

    let prefix = '';
    if ((typeof urlPrefix === 'string') && (urlPrefix.trim() !== '')) {
        prefix = urlPrefix;
        if (prefix.substr(0, 1) === '/') {
            prefix = prefix.substr(1);
        }
        if (prefix.substr(-1, 1) !== '/') {
            prefix = prefix + '/';
        }
    }

    /**
     * Log to logger if set in options
     * @param {string} logMessage Log message
     * @param {string} logLevel Log level, default is debug
     */
    function log(logMessage, logLevel = 'debug') {
        if (typeof options.logger !== 'undefined') {
            options.logger.log({
                level: logLevel,
                message: logMessage
            });
        }
    }

    /**
     * Get the inventory of a publisher and retry every 5 minutes, if not present.
     * @param {string} publisher Name of the publisher.
     * @param {function} inventoryCallback Callback function with the inventory as parameter.
    async function getInventory(publisher, inventoryCallback, inventoryOptions) {
     * @param {object} options Additional options like time span, prefix and id.
     * @returns {Promise} Promise for an inventory.
     */
    async function getInventory(publisher, inventoryCallback, inventoryOptions) {
        return new Promise(function (resolve, reject) {
            function receiveInventory() {
                let bodyData = '';
                let url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/inventory/';
                if (typeof inventoryOptions !== 'undefined') {
                    if (typeof inventoryOptions.prefix !== 'undefined') {
                        url = url + inventoryOptions.prefix + '/';
                    }
                    url = url + publisher;
                    if (typeof inventoryOptions.field !== 'undefined') {
                        url = url + '/' + inventoryOptions.field;
                    }
                    if (typeof inventoryOptions.id !== 'undefined') {
                        url = url + '/' + inventoryOptions.id;
                    }
                    if ((typeof inventoryOptions.from === 'object') && (typeof inventoryOptions.from.toISOString === 'function')) {
                        url = url + '/from/' + inventoryOptions.from.toISOString();
                    }
                    if ((typeof inventoryOptions.to === 'object') && (typeof inventoryOptions.to.toISOString === 'function')) {
                        url = url + '/to/' + inventoryOptions.to.toISOString();
                    }
                    if ((typeof inventoryOptions.between === 'object') && Array.isArray(inventoryOptions.between) && (inventoryOptions.between.length === 2) && (typeof inventoryOptions.between[0].toISOString === 'function') && (typeof inventoryOptions.between[1].toISOString === 'function')) {
                        url = url + '/between/' + inventoryOptions.between[0].toISOString() + '/' + inventoryOptions.between[1].toISOString();
                    }
                    if (typeof inventoryOptions.suffix !== 'undefined') {
                        url = url + inventoryOptions.suffix;
                    }
                } else {
                    url = url + publisher;
                }
                log('Get inventory from ' + url);
                request({
                    url,
                    headers: {
                        "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                    },
                    method: 'GET'
                }).on('data', function (data) {
                    bodyData += data;
                }).on('error', function (e) {
                    // Fehler abfangen
                    log(e, 'error');
                }).on('response', function (resp) {
                    if (resp.statusCode === 404) {
                        log('Publisher ' + publisher + ' not connected to deliver inventory -> try again in 5 minutes.', 'warn');
                        setTimeout(receiveInventory, 300000);
                    }
                }).on('end', function () {
                    log('Inventory received');
                    if (typeof bodyData === 'string') {
                        try {
                            bodyData = JSON.parse(bodyData);
                        } catch (err) {
                            log('Error in inventory response: ' + bodyData + ' (' + JSON.stringify(err) + ')', 'error');
                        }
                    }
                    inventoryCallback(bodyData);
                    resolve(bodyData);
                });
            }
            receiveInventory();
        });
    }

    /**
     * Register subscription via websocket for the notifications of a publisher.
     * @param {string} publisher Name of the publisher.
     * @param {function} notificationCallback Callback function for notification calls. It should take a parameter for the method and a second one for the data.
     * @returns {object} Returns an object with the function endSubscription.
     */
    function registerWebsocket(publisher, notificationCallback) {
        let isActive = false;
        const url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/websocket/' + publisher;
        log('Connect to websocket ' + url);
        const socket = require('socket.io-client')(url);
        socket.on('connect', () => {
            socket.emit('authentication', {
                username: options.credentials.username,
                password: options.credentials.password
            });
        });
        socket.on('connect', function () {
            isActive = true;
        });
        socket.on('PUT', function (data) {
            if (typeof data === 'string') {
                data = JSON.parse(data);
            }
            notificationCallback('PUT', data);
        });
        socket.on('POST', function (data) {
            if (typeof data === 'string') {
                data = JSON.parse(data);
            }
            notificationCallback('POST', data);
        });
        socket.on('DELETE', function (data) {
            notificationCallback('DELETE', data);
        });
        socket.on('disconnect', function () {
            isActive = false;
        });
        return {
            endSubscription: function () {
                socket.close();
            },
            isActive
        }
    }

    /**
     * Register subscription for the notifications of a publisher.
     * @param {string} publisher Name of the publisher.
     * @param {function} notificationCallback Callback function for notification calls. It should take a parameter for the method and a second one for the data.
     * @param {object} subscriptionOptions Additional options like time span and prefix.
     * @returns {object} Returns an object with the function endSubscription.
     */
    function registerSubscription(publisher, notificationCallback, subscriptionOptions) {

        /** Is subsctiption active? */
        let activeSubscription = false;

        // Define endpoint for REST notification callback
        const restPath = "/ham-dex-subscriber/notification/" + publisher.replace(/\s/g, '-');
        const restCompletePath = "/" + prefix + "ham-dex-subscriber/notification/" + publisher.replace(/\s/g, '-');
        log('Add rest endpoint ' + restPath);
        app.use(restPath, function (req, res, next) {
            let body = (typeof req.body !== 'undefined') ? req.body : null;
            if (req.originalUrl != restCompletePath) {
                let suffix = req.originalUrl.substring(restCompletePath.length + 1);
                if (suffix != '') {
                    body = suffix;
                }
            }
            if (activeSubscription) {
                notificationCallback(req.method, body);
                res.json(true);
            } else {
                next();
            }
        });

        // Register subscription
        const restPathSubscription = ((typeof subscriptionOptions === 'undefined') || (typeof subscriptionOptions.beginningSlash !== 'boolean') || subscriptionOptions.beginningSlash ? '/' : '') + prefix + "ham-dex-subscriber/notification/" + publisher.replace(/\s/g, '-');
        let subscriptionId = null;
        let bodyData = '';
        url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/subscription';
        let subscriptionRequest = {
            "name": options.subscriber,
            "description": options.subscriber + ' - ' + publisher,
            "protocol": (typeof options.callback.protocol !== 'string') ? "HTTP" : options.callback.protocol.toUpperCase(),
            "serviceType": "REST",
            "format": "JSON",
            "communication": "SYNCHRONOUS",
            "channel": "Default",
            "filterProfile": {
                "name": publisher
            },
            "subscriber": {
                "name": options.subscriber
            },
            "host": options.callback.host,
            "port": options.callback.port,
            "path": restPathSubscription,
            "persist": false
        };
        if (typeof subscriptionOptions !== 'undefined') {
            if ((typeof subscriptionOptions.from === 'object') && (typeof subscriptionOptions.from.toISOString === 'function')) {
                subscriptionRequest.filterFrom = subscriptionOptions.from.toISOString();
            }
            if ((typeof subscriptionOptions.to === 'object') && (typeof subscriptionOptions.to.toISOString === 'function')) {
                subscriptionRequest.filterTo = subscriptionOptions.to.toISOString();
            }
        }
        if (typeof options.id) {
            subscriptionRequest.subscriber.id = options.id;
        }
        log('Register Subscription ' + JSON.stringify(subscriptionOptions));
        request({
            url,
            headers: {
                "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
            },
            method: 'POST',
            json: subscriptionRequest
        }).on('error', function (e) {
            // Log error to console
            log(e, 'error');
        }).on('data', function (data) {
            bodyData += data;
        }).on('response', function (resp) { }).on('end', function () {
            try {
                const response = JSON.parse(bodyData);
                if ((typeof response !== 'undefined') && (typeof response.id !== 'undefined')) {
                    subscriptionId = response.id;
                    log('Subscription successful ' + subscriptionId);
                } else {
                    log('Subscription not successful: ' + bodyData);
                }
                activeSubscription = true;
            } catch (exception) {
                log('Error in subscription response: ' + bodyData + ' (' + JSON.stringify(exception) + ')', 'error');
            }
        });

        // End subscription
        const endSubscription = async function () {
            return new Promise(function (resolve, reject) {
                activeSubscription = false;
                let bodyData = '';
                if (subscriptionId !== null) {
                    log('End subscription ' + subscriptionId);
                    request({
                        url: ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/subscription/' + subscriptionId,
                        headers: {
                            "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                        },
                        method: 'DELETE'
                    }).on('error', function (e) {
                        // Log error to console
                        log(e, 'error');
                        reject(e);
                    }).on('data', function (data) {
                        bodyData += data;
                    }).on('end', function () {
                        resolve(bodyData);
                    });
                }
            });
        }

        /**
         * @returns True, if this subscription is active.
         */
        const isActive = function () {
            return new Promise(function (resolve, reject) {
                if (subscriptionId !== null) {
                    log('Check subscription ' + subscriptionId);
                    request({
                        url: ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/subscription/' + subscriptionId,
                        headers: {
                            "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                        },
                        method: 'GET'
                    }).on('error', function (e) {
                        // Log error to console
                        log(e, 'error');
                        reject(e);
                    }).on('response', function (response) {
                        activeSubscription = (response.statusCode == 200);
                        resolve(activeSubscription);
                    });
                }
            });
        }

        return {
            endSubscription: endSubscription,
            isActive: isActive
        };
    }

    /**
     * Call hateway api
     * @param {*} publisher Name of the publisher.
     * @param {*} method Http method.
     * @param {*} api Name or path of the api.
     * @param {*} data Data to send to gateway api.
     * @param {*} callback Callback function for the gateway answer. It should take a parameter for the data.
     */
    const callGateway = async function (publisher, method, api, data, callback) {
        return new Promise(function (resolve, reject) {
            let bodyData = '';
            request({
                url: ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/gateway/' + publisher + '/' + api,
                headers: {
                    "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                },
                method: method,
                json: data
            }).on('error', function (e) {
                // Log error to console
                log(e, 'error');
            }).on('data', function (data) {
                bodyData += data;
            }).on('end', function () {
                if (typeof callback !== 'undefined') {
                    if (typeof bodyData === 'string') {
                        bodyData = JSON.parse(bodyData);
                    }
                    callback(bodyData);
                }
                resolve(bodyData);
            });
        });
    }

    /**
     * Send data as post to input service of a publisher.
     * @param {string} publisher Name of the publisher.
     * @param {*} data Data to send to input service.
     * @param {object} options Additional options like time span, prefix and id.
     */
    const sendInputPost = async function (publisher, data, inputOptions) {
        return new Promise(function (resolve, reject) {
            let bodyData = '';
            let url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/input/';
            if ((typeof inputOptions !== 'undefined') && (typeof inputOptions.prefix !== 'undefined')) {
                url = url + inputOptions.prefix + '/';
            }
            url = url + publisher;
            if (typeof inputOptions.suffix !== 'undefined') {
                url = url + inputOptions.suffix;
            }
            request({
                url: url,
                headers: {
                    "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                },
                method: 'POST',
                json: data
            }).on('error', function (e) {
                // Log error to console
                log(e, 'error');
            }).on('data', function (data) {
                bodyData += data;
            }).on('end', function () {
                resolve(bodyData);
            });
        });
    }

    /**
     * Send data as put to input service of a publisher.
     * @param {string} publisher Name of the publisher.
     * @param {*} data Data to send to input service.
     * @param {object} options Additional options like time span, prefix and id.
     */
    const sendInputPut = async function (publisher, data, inputOptions) {
        return new Promise(function (resolve, reject) {
            let bodyData = '';
            let url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/input/';
            if ((typeof inputOptions !== 'undefined') && (typeof inputOptions.prefix !== 'undefined')) {
                url = url + inputOptions.prefix + '/';
            }
            url = url + publisher;
            if (typeof inputOptions.suffix !== 'undefined') {
                url = url + inputOptions.suffix;
            }
            request({
                url: url,
                headers: {
                    "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                },
                method: 'PUT',
                json: data
            }).on('error', function (e) {
                // Log error to console
                log(e, 'error');
            }).on('data', function (data) {
                bodyData += data;
            }).on('end', function () {
                resolve(bodyData);
            });
        });
    }

    /**
     * Delete data by an input service of a publisher.
     * @param {string} publisher Name of the publisher.
     * @param {*} id ID of the data.
     * @param {object} options Additional options like time span, prefix and id.
     */
    const sendInputDelete = async function (publisher, id, inputOptions) {
        return new Promise(function (resolve, reject) {
            let bodyData = '';
            let url = ((typeof options.esb.protocol === 'undefined') ? 'https' : options.esb.protocol) + '://' + options.esb.host + ":" + options.esb.port + '/rest/input/';
            if ((typeof inputOptions !== 'undefined') && (typeof inputOptions.prefix !== 'undefined')) {
                url = url + inputOptions.prefix + '/';
            }
            url = url + publisher;
            if (typeof inputOptions.suffix !== 'undefined') {
                url = url + inputOptions.suffix;
            }
            url = url + '/' + id;
            request({
                url: url,
                headers: {
                    "Authorization": "Basic " + new Buffer(options.credentials.username + ":" + options.credentials.password).toString("base64")
                },
                method: 'DELETE'
            }).on('error', function (e) {
                // Log error to console
                log(e, 'error');
            }).on('data', function (data) {
                bodyData += data;
            }).on('end', function () {
                resolve(bodyData);
            });
        });
    }

    return {
        getInventory: getInventory,
        registerSubscription: registerSubscription,
        registerWebsocket: registerWebsocket,
        callGateway: callGateway,
        input: {
            post: sendInputPost,
            put: sendInputPut,
            delete: sendInputDelete
        }
    }
};